/*
 * IntuitionBuildGraph.qml  -  Construct 10 graphs 
 *
 * Copyright © 2023,2024 IntuitionSailing.com.  All rights reserved.
 * Contact: IntiutionSailing http://www.intuitionsailing.com don@intuitionsailing.com
 *
 * IntuitionGraphs defines 1 to 10 IntuitionBuildGraph objects which displays each graph.
 * Each graph displays 1 to 10 lines.  Reads dbus values every graphInterval and updates
 * the y[index][] with dbus data and refresh the graph.  RPN formulas are based on the
 * dbus element values, e.g., [0] first.
 *
*/

import QtQuick 1.1
import com.victron.velib 1.0
import "utils.js" as Utils

Item {
  id: buildGraph

  property int       graphId

  property string    victronMappingPrefix: "com.victronenergy.system/ServiceMapping/"
  property string    servicePrefix: "com.victronenergy."
  property string    mappingPrefix: "com_victronenergy_"
  property string    myPrefix: "com.victronenergy.settings/Settings/Intuition"
  property int       chartBorderWidth: 1
  property color     defaultBorderColor: "#808080"
  property int       chartLineWidth: 2
  property string    commas: ",,,,,,,,,"
  property int       graphPoints: 48

  property bool      firstTime: true
  property string    refreshTime                    // Display next refresh time
  property int       yAxisDecimals: getDecimals(0)  // Decimals on labels y axis scale

  property real      yMin: 0
  property real      yMax: 0
  property real      yAvg: 0
  property real      pZero: 0  // Relative position of zero

  property VBusItem  graphIntervalItem: VBusItem {bind: Utils.path(myPrefix,"/Graphs/",graphId,"/Interval") }
  property int       graphInterval: (graphIntervalItem.valid && graphIntervalItem.value != "" ? 
    graphIntervalItem.value : 24 ) * 60 * 60 * 1000 / graphPoints
  property int       graphCounter: 0
  property int       ticks: graphInterval * graphPoints / 1000 / 3600

  property color     defaultBodyColor: "#faf9f5"
  property VBusItem  bodyColorDbus: VBusItem {bind: Utils.path(myPrefix,"/Graphs/",graphId,"/BodyColor")}
  property color     bodyColor : (bodyColorDbus.valid && bodyColorDbus.value != "" ? 
    bodyColorDbus.value : defaultBodyColor)
  property int       fadeColorPercent: 0 // 25

  property VBusItem  borderColorDbus: VBusItem {bind: Utils.path(myPrefix,"/Graphs/",graphId,"/BorderColor")}
  property color     borderColor: (borderColorDbus.valid && borderColorDbus.value != "" ? 
    borderColorDbus.value : defaultBorderColor)

  property VBusItem  chartTypeDbus: VBusItem {bind: Utils.path(myPrefix,"/Graphs/",graphId,"/Type")}
  property int       chartType: (chartTypeDbus.valid && chartTypeDbus != "" ? parseInt(chartTypeDbus.value) : 0 )

  property VBusItem  captionsDbus : VBusItem {bind: Utils.path(myPrefix,"/Graphs/",graphId,"/Captions")}
  property variant   captionsArray: (captionsDbus.valid ? captionsDbus.value : "").split(",")
  property int       captionsArrayLength: captionsArray.length              // Used everywhere

  property bool      isVisible :  (captionsArray != "" ? true : false)

  property VBusItem  captionFontSizeDbus : VBusItem {bind: Utils.path(myPrefix,"/Graphs/",graphId,"/CaptionsFontSize")}
  property int       captionsFontSize: (captionFontSizeDbus.valid && captionFontSizeDbus.value != "" ? parseInt(captionFontSizeDbus.value) : 10 )

  property string    defaultLineColors: "#e10000,#075f07,#0e2a53,#e100e1,#e17014,#00e1e1,#80000,#c0c0c0,#808080,#000000"
  property variant   defaultLineColorArray : defaultLineColors.split(",")
  property VBusItem  lineColorsDbus : VBusItem {bind: Utils.path(myPrefix,"/Graphs/",graphId,"/LineColors")}
  property string    lineColors: (lineColorsDbus.valid && lineColorsDbus.value != "" ?  lineColorsDbus.value  + "," + defaultLineColors : defaultLineColors)
  property variant   lineColorArray: lineColors.split(",")

  // Line Types: 0 = solid line, 1 = dash line
  property string    defaultLineTypes: "0,0,0,0,0,0,0,0,0,0"
  property VBusItem  lineTypesDbus : VBusItem {bind: Utils.path(myPrefix,"/Graphs/",graphId,"/LineTypes")}
  property string    lineTypes: (lineTypesDbus.valid && lineTypesDbus.value != "" ?  lineTypesDbus.value : defaultLineTypes)
  property variant   lineTypesArray: lineTypes.split(",")

  property string    defaultBackgroundColor: "#d0d0d0"
  property VBusItem  backgroundColorsDbus : VBusItem {bind: Utils.path(myPrefix,"/Graphs/", graphId, "/BackgroundColor")}  
  property string    backgroundColor: (backgroundColorsDbus.valid && backgroundColorsDbus.value != "" ? 
     backgroundColorsDbus.value : defaultBackgroundColor )

  property string    defaultDecimals: "0,0,0,0,0,0,0,0,0,0"
  property VBusItem  decimalsDbus : VBusItem {bind: Utils.path(myPrefix,"/Graphs/",graphId,"/Decimals")}
  property variant   decimals: (decimalsDbus.valid && decimalsDbus.value != "" ? decimalsDbus.value : defaultDecimals)
  property variant   decimalsArray: decimals.split(",")

  // defaultYMin & defaultYMax sets the Y axis scale.  Used for fixed ranges like SOC: 0-100%
  property VBusItem  defaultYMin: VBusItem {bind: Utils.path(myPrefix, "/Graphs/",graphId,"/YMin")}
  property VBusItem  defaultYMax: VBusItem {bind: Utils.path(myPrefix, "/Graphs/",graphId,"/YMax")}

  // dbus cache data, string ","
  property VBusItem  graphCache : VBusItem {bind: Utils.path(myPrefix,"/Graph/Cache",graphId)}

  // Placeholder for 48 data points for my graphs, up to 10 lines (including totals and formulas)
  property list<Path> pathList: [
    Path {
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
    },
    Path {
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
    },
    Path {
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
    },
    Path {
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
    },
    Path {
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
    },
    Path {
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
    },
    Path {
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
    },
    Path {
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
    },
    Path {
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
    },
    Path {
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
      PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{} PathLine{}
    }
  ]

  // To fetch values from Victron dbus - first split services csv and service mapings into 
  // arrays servicesArray and serviceMappingArrayList. Using the service mappings, lookup 
  // the service (results are stored in serviceMappingArrayTmp).  If services[i] is defined
  // the set _services[i] = servicesArray[i]; else use the results of serviceMappingArrayTmp[i].value.
  // Examples:
  //   Services: tank.mopeka_eff299c31e5d
  //   Service Mappings: system, temperature_20, battery_288...
  property VBusItem  servicesDbus : VBusItem {bind: Utils.path(myPrefix,"/Graphs/",graphId,"/Services")}

  property string    services: (servicesDbus.valid ? servicesDbus.value : "")   
  property VBusItem  serviceMappingsDbus : VBusItem {bind: Utils.path(myPrefix,"/Graphs/",graphId,"/ServiceMappings")}
  property string    serviceMappings:  (serviceMappingsDbus.valid ? serviceMappingsDbus.value : "")
  
  // If instance starts with '/', then dbus instance; otherwise RPN formula.
  // Examples: /Pressure, /Temperature, /Dc/0/Voltage, or RPN formula "[1] [4] /", where [1],[4] reference 
  // instance value
  property VBusItem  instancesDbus : VBusItem {bind: Utils.path(myPrefix,"/Graphs/",graphId,"/Instances")}
  property string    instances: (instancesDbus.valid ? instancesDbus.value : "")
  property variant   instancesArray: (instances + commas).split(",")  
  
  property variant   servicesArray: (services + commas).split(",")
  
  property variant   serviceMappingArray: (serviceMappings + commas).split(",")

  // *** Combining services and service mappings throws dbus invalid object/path error ***
  // *** For now, keep service and service mappings separate graph setup values ***
  // *** Note, should be able to combine service mapping[i] + service[i]
  // *** Verify error handling in Victron code
  property list<VBusItem> serviceMappingArrayList : [
    VBusItem { bind: Utils.path(victronMappingPrefix, mappingPrefix , serviceMappingArray[0])},
    VBusItem { bind: Utils.path(victronMappingPrefix, mappingPrefix , serviceMappingArray[1])},
    VBusItem { bind: Utils.path(victronMappingPrefix, mappingPrefix , serviceMappingArray[2])},
    VBusItem { bind: Utils.path(victronMappingPrefix, mappingPrefix , serviceMappingArray[3])},
    VBusItem { bind: Utils.path(victronMappingPrefix, mappingPrefix , serviceMappingArray[4])},
    VBusItem { bind: Utils.path(victronMappingPrefix, mappingPrefix , serviceMappingArray[5])},
    VBusItem { bind: Utils.path(victronMappingPrefix, mappingPrefix , serviceMappingArray[6])},
    VBusItem { bind: Utils.path(victronMappingPrefix, mappingPrefix , serviceMappingArray[7])},
    VBusItem { bind: Utils.path(victronMappingPrefix, mappingPrefix , serviceMappingArray[8])},
    VBusItem { bind: Utils.path(victronMappingPrefix, mappingPrefix , serviceMappingArray[9])}
  ]

  property variant _services : [
    (serviceMappingArrayList[0].valid && serviceMappingArrayList[0].value != "" ? serviceMappingArrayList[0].value : servicePrefix + servicesArray[0]),
    (serviceMappingArrayList[1].valid && serviceMappingArrayList[1].value != "" ? serviceMappingArrayList[1].value : servicePrefix + servicesArray[1]),
    (serviceMappingArrayList[2].valid && serviceMappingArrayList[2].value != "" ? serviceMappingArrayList[2].value : servicePrefix + servicesArray[2]),
    (serviceMappingArrayList[3].valid && serviceMappingArrayList[3].value != "" ? serviceMappingArrayList[3].value : servicePrefix + servicesArray[3]),
    (serviceMappingArrayList[4].valid && serviceMappingArrayList[4].value != "" ? serviceMappingArrayList[4].value : servicePrefix + servicesArray[4]),
    (serviceMappingArrayList[0].valid && serviceMappingArrayList[5].value != "" ? serviceMappingArrayList[5].value : servicePrefix + servicesArray[5]),
    (serviceMappingArrayList[1].valid && serviceMappingArrayList[6].value != "" ? serviceMappingArrayList[6].value : servicePrefix + servicesArray[6]),
    (serviceMappingArrayList[2].valid && serviceMappingArrayList[7].value != "" ? serviceMappingArrayList[7].value : servicePrefix + servicesArray[7]),
    (serviceMappingArrayList[3].valid && serviceMappingArrayList[8].value != "" ? serviceMappingArrayList[8].value : servicePrefix + servicesArray[8]),
    (serviceMappingArrayList[4].valid && serviceMappingArrayList[9].value != "" ? serviceMappingArrayList[9].value : servicePrefix + servicesArray[9])
  ]

  // dbus value list using path/element(s)
  property list<VBusItem> valueList: [
  VBusItem { bind: Utils.path(_services[0], instancesArray[0]); onValueChanged: copyCurrentValues() },
  VBusItem { bind: Utils.path(_services[1], instancesArray[1]); onValueChanged: copyCurrentValues() },
  VBusItem { bind: Utils.path(_services[2], instancesArray[2]); onValueChanged: copyCurrentValues() },
  VBusItem { bind: Utils.path(_services[3], instancesArray[3]); onValueChanged: copyCurrentValues() },
  VBusItem { bind: Utils.path(_services[4], instancesArray[4]); onValueChanged: copyCurrentValues() },
  VBusItem { bind: Utils.path(_services[5], instancesArray[5]); onValueChanged: copyCurrentValues() },
  VBusItem { bind: Utils.path(_services[6], instancesArray[6]); onValueChanged: copyCurrentValues() },
  VBusItem { bind: Utils.path(_services[7], instancesArray[7]); onValueChanged: copyCurrentValues() },
  VBusItem { bind: Utils.path(_services[8], instancesArray[8]); onValueChanged: copyCurrentValues() },
  VBusItem { bind: Utils.path(_services[9], instancesArray[9]); onValueChanged: copyCurrentValues() }
  ]

  // Note, we can't update global variant arrays, so we still need these
  // goofey curValY0,curValY1... values.  Use copyCurrentValues() to set
  property real    curValY0: 0
  property real    curValY1: 0
  property real    curValY2: 0
  property real    curValY3: 0
  property real    curValY4: 0
  property real    curValY5: 0
  property real    curValY6: 0
  property real    curValY7: 0
  property real    curValY8: 0
  property real    curValY9: 0

  // If the caption is '*', then use the name defined by Cerbo Gx
  property list <VBusItem> customNameList: [ 
    VBusItem { bind: Utils.path(_services[0], "/CustomName") },
    VBusItem { bind: Utils.path(_services[1], "/CustomName") },
    VBusItem { bind: Utils.path(_services[2], "/CustomName") },
    VBusItem { bind: Utils.path(_services[3], "/CustomName") },
    VBusItem { bind: Utils.path(_services[4], "/CustomName") },
    VBusItem { bind: Utils.path(_services[5], "/CustomName") },
    VBusItem { bind: Utils.path(_services[6], "/CustomName") },
    VBusItem { bind: Utils.path(_services[7], "/CustomName") },
    VBusItem { bind: Utils.path(_services[8], "/CustomName") },
    VBusItem { bind: Utils.path(_services[9], "/CustomName") }
  ]

  // Common units of measure (UOM), for value display 
  // Verify if Victron has these UOMs defined? 
  property variant displayUOM: { 
    "/Level":"%",
    "/Temperature":"°F",
    "/Pressure":"hPa", 
    "/Humidity":"%",
    "/Soc":"%",
    "/Dc/Battery/Soc":"%",
    "/Dc/Pv/Current":"A",
    "/Dc/Pv/Power":"W",
    "/Dc/System/Power":"W",
    "/Dc/0/Voltage":"V",
    "/Dc/0/Current":"A",
    "/Dc/Battery/Current":"A",
    "/Dc/Battery/Voltage":"V",
    "/Dc/Battery/Power":"W",
    "/Dc/Pv/Current":"A",
    "/Dc/0/Power":"W",
    "/Dc/0/Temperature":"°F",
    "/Dc/1/Voltage":"V",
    "/Dc/Battery/ConsumedAmphours":"Ah",
    "/ConsumedAmphours":"Ah",
    "/TimeToGo":"TTG",
    "/Pv/V":"V", 
    "/Pv/0/V":"V",
    "/Pv/1/V":"V",
    "/Pv/P":"W",
    "/Pv/0/P":"W",
    "/Pv/1/P":"W",
    "/Dc/Pv/Power":"W",
    "/Yield/Power":"W",
    "/History/Daily/0/Consumption":"kWh",
    "/History/Daily/0/Yield":"kWh",
    "/History/Daily/1/Consumption":"kWh",
    "/History/Daily/1/Yield":"kWh"
    }

  //
  // copyCurrentValues() - copy current values into curValYn
  //
  function copyCurrentValues() {

  if(firstTime) return
  
  // Store current values in array for convenience
  curValY0 = (isValue(0) ? cToF(instancesArray[0], valueList[0].value) : 0)
  curValY1 = (isValue(1) ? cToF(instancesArray[1], valueList[1].value) : 0)
  curValY2 = (isValue(2) ? cToF(instancesArray[2], valueList[2].value) : 0)
  curValY3 = (isValue(3) ? cToF(instancesArray[3], valueList[3].value) : 0)
  curValY4 = (isValue(4) ? cToF(instancesArray[4], valueList[4].value) : 0)
  curValY5 = (isValue(5) ? cToF(instancesArray[5], valueList[5].value) : 0)
  curValY6 = (isValue(6) ? cToF(instancesArray[6], valueList[6].value) : 0)
  curValY7 = (isValue(7) ? cToF(instancesArray[7], valueList[7].value) : 0)
  curValY8 = (isValue(8) ? cToF(instancesArray[8], valueList[8].value) : 0)
  curValY9 = (isValue(9) ? cToF(instancesArray[9], valueList[9].value) : 0)

  // if a value changed, recalculate formulas
  doRPN()

  }

  //
  // doRPN() - RPN calculations
  //
  function doRPN() {

    // If first time, instance value (/Temperature)
    if(firstTime) return

    var tmpCurrentValues = [curValY0, curValY1, curValY2, curValY3, curValY4, curValY5, curValY6, curValY7, curValY8, curValY9]

    // Calculate rpn formula, 1 2 + = 3,  [0] [1] + == value 1 + value 2
    // Note, use space to delineate the value ([n]=dbus value, number, or operand)
    for(var i=0; i<captionsArrayLength; i++) {

      // If we have a formula
      if( isFormula(i) && instancesArray[i].length > 0) {

        var rpn = instancesArray[i].split(" ")
        var stack = [0,0,0,0]

        if(rpn.length > 0) {

          for(var j=0; j<rpn.length; j++) {

            //if dbus value or constant, push onto stack
            if(startsWith(rpn[j],"[.0123456789")) {
              // Push down stack values
              for(var k=stack.length-2; k>=0; k--) {
                stack[k+1] = stack[k]
              }
              // Save rpn value
              //stack[0] = rpn[j]
              stack[0] = stackToValue(rpn[j],tmpCurrentValues)
            }  // startsWith

            else {    // otherwise, operator + - * /

              // Pop stack values, fetch 2 values or constants from stack
              var value = stack[0]
              for(var k=1  ;k<stack.length ; k++) {
                stack[k-1] = stack[k]
              }
              
              // Do the calculation
              switch (rpn[j]) {
                case "+": 
                  stack[0] = stack[0] + value
                  break
                case "-": 
                  stack[0] = stack[0] - value
                  break
                case "*": 
                  stack[0] = stack[0] * value
                  break
                case "/": 
                  stack[0] = stack[0] / value
              } // switch

            } // startsWith [.0123456789
          }  //j (rpn.length)

          // Store the result
          //tmpCurrentValues[i] = (typeof stack[0]=="number" ? stack[0] : parseFloat(stack[0]))
          tmpCurrentValues[i] = stack[0]

          // Store the calculatead values in corrsespondingd curValYx
          if(i==0) curValY0 = tmpCurrentValues[0]
          if(i==1) curValY1 = tmpCurrentValues[1]
          if(i==2) curValY2 = tmpCurrentValues[2]
          if(i==3) curValY3 = tmpCurrentValues[3]
          if(i==4) curValY4 = tmpCurrentValues[4]
          if(i==5) curValY5 = tmpCurrentValues[5]
          if(i==6) curValY6 = tmpCurrentValues[6]
          if(i==7) curValY7 = tmpCurrentValues[7]
          if(i==8) curValY8 = tmpCurrentValues[8]
          if(i==9) curValY9 = tmpCurrentValues[9]
        } // if rpn

      } // if isFormula(i).length > 0
      
    } // i

//if(graphId==9) console.log("doRPN ...",curValY0, curValY1, curValY2)

  }

  //
  // startsWith
  // example: string="[1] 2 3 + +"  values = "[ 0 1 2 3 .... 9]"
  //
  function startsWith(string, values) {
    for(var i=0; i<values.length; i++) {
      if(string.substr(0,1) == values.substr(i,1)) return true
    }
    return false
  }

  //
  // stackToValue(s)
  // s is either a dbus value "[0]" or number "123"
  //
  function stackToValue(s,v) {
    if(typeof s == "number") return s  // already type conversion
    if(s.substr(0,1) == "[") {
      var i = parseInt(s.substr(1,1))
      return v[i]
    }
    else
      return parseFloat(s)
  }

  //
  // isForumla - helper function
  //
  function isFormula(index) {
    return (startsWith(instancesArray[index],"/") ? false : true)
  }

  // isValue -- helper function
  //
  function isValue(index) {
    return (valueList[index].valid && valueList[index].value != "")
  }

  //
  // fadeColor(color, percent)
  //
  function fadeColor(color, percent) {

    if(color[0] != "#") return color

    // https://gist.github.com/renancouto/4675192
    var num = parseInt(color.replace("#",""),16)
    var amt = Math.round(2.55 * percent)
    var R = (num >> 16) + amt
    var B = (num >> 8 & 0x00FF) + amt
    var G = (num & 0x0000FF) + amt

    return "#" + (0x1000000 + 
        (R<255?R<1?0:R:255)*0x10000 + 
        (B<255?B<1?0:B:255)*0x100 +
        (G<255?G<1?0:G:255)).toString(16).slice(1)
  }


  //
  // getColor(index, percent) - helper function for graph lines, captions, etc.
  //
  function getColor(index, percent, control) {

    var color = lineColorArray[index]
//console.log("******",graphId, control.y , pZero, control.y - pZero)

    // Sanity check, starts with '#'
    if(color[0] != "#") return defaultLineColorArray[0] 

    // Check if positive/negative colors
    var theColors = color.split("|")

    // Check if valid, total line or QT color; return
    if(isValue(index) || isFormula(index)) {
//console.log("I am here",control)
      if(theColors.length > 1)                   // Positive/negative value
        return (control.y - pZero <= 0 ? theColors[1] : theColors[0])
//        return (valueList[index].value >=0 ? theColors[0] : theColors[1])
      else
        return theColors[0]
    }

    // Not valid value, fade colors
    if(theColors.length > 1)
      return (valueList[index].value >=0 ? fadeColor(theColors[0], percent) : fadeColor(theColors[1], percent))

    // In all other cases, fade color
    return fadeColor(color, percent)
  }

  
  //
  // getDecimals(i) - helper function for display decimals
  //
  function getDecimals(i) {
    return (i < decimalsArray.length ? parseInt(decimalsArray[i]) : 0)
  }

  //
  // cToFHelper  -- if Temperature then convert to F; else return value
  //
  function cToFHelper(instance, value ) {
    //return (instance.indexOf("/Temperature") >= 0 ? (value * 9 / 5 + 32) : value )
    return (instance == "/Temperature" ? (value * 9 / 5 + 32) : value )
  }

  //
  // cToF  -- if Temperature then convert to F; else return value
  // with set decimals
  //
  function cToF(instance, value , decimals) {

    if(value === undefined) return "--"

    return (decimals === undefined ? cToFHelper(instance, value) :
      cToFHelper(instance, value).toFixed(decimals) )
  }

  //
  // scaleYAxis - function to scale the Y axis, set yMin, yMax, and yAvg
  //
  function scaleYAxis(y) {

    // Get min/max for scaleYAxis --- at this point we should have all the y[][] values, including the totals
    for(var i=0; i< captionsArrayLength; i++) {
      if( (isValue(i) || isFormula(i) ) && captionsArray[i] != "" && captionsArray[i] != "-") {
        for(var j=0; j< graphPoints; j++) {
          yMin = (defaultYMin.valid && defaultYMin.value != "" ? 
            defaultYMin.value : 
            (y[i][j] < yMin ? y[i][j] : yMin)
          )
          yMax = (defaultYMax.valid && defaultYMax.value != "" ? 
            defaultYMax.value :
            (yMax < y[i][j] ? y[i][j] : yMax)
          )
        }
      }
    }

    // Add small separation in values
    if(yMax == yMin) 
      yMax = yMax + 1
    yAvg = (yMax + yMin) / 2

    return -1 * (chartFrame.height - 2 * chartFrame.border.width - 2) / (yMax - yMin)
  }

  //
  // writeDbusCache - the values are stored in a ',' deliminated string
  //
  function writeDbusCache(y) {
    var dbusString = ""
    for(var i = 0; i < captionsArrayLength; i++) {
      for(var j = 0; j < graphPoints; j++) {
        dbusString += (i==0 && j==0 ? "" : ",") + (y[i][j] === undefined ? 0 : y[i][j].toFixed(getDecimals(i) + 2))
      }
    }
    // Update dbus cache
    graphCache.setValue(dbusString)

  }

  //
  // displayInterval()  - Helper function to format the graph display interval,
  // top right
  //
  function displayInterval() {

    // Interval
    var i = graphInterval / 1000      // seconds

//    var n = Qt.formatDateTime(new Date(new Date().getTime() + graphInterval), "h:mm:ss")

    var s = i.toFixed(1) + "sec"
    if (i > 60)
      {
      i =  i / 60
      s = i.toFixed(1) + "min"
    }
    if(i > 60)
    {
      i = i / 60
      s = i.toFixed(1) + "hr"
    }

    // Graph
    var t = graphInterval / 60 / 60 / 1000 * graphPoints
    if(t >= 1) return t.toFixed(1) + "hr @ " + s
    t=t*60
    if(t >=1 ) return t.toFixed(1) + "min @ " + s
    t=t*60
    return t.toFixed(1) + "sec @ " + s
  }

  //
  // makeCaption(index)
  //
  function makeCaption(index) {

    var caption = ""

    // Title
    if(captionsArray[index] == "*") 
      caption += customNameList[index].value
    else if(captionsArray[index] != "~" && captionsArray[index] != "-" && captionsArray[index] != "%") 
      caption += captionsArray[index].replace("~","").replace("-","")

    // Display Solar /State: Bulk, Abs, Float
    if(instancesArray[index] == "/State") {
      caption += ["(Off)","","(Fa)","(B)","(A)","(Fl)","(S)","(Eq)"][valueList[index].value]
    }
    else {
      // Add space between name and value
      caption += (caption.length > 0 ? " " : "")

      // Store current values in tmpCurrentValues for convenience
      var tmpCurrentValues = [curValY0, curValY1, curValY2, curValY3, curValY4, curValY5, curValY6, curValY7, curValY8, curValY9]

      if(captionsArray[index] != "%" ) {
        caption += tmpCurrentValues[index].toFixed( getDecimals(index) )
      }

      // UOM
      if(captionsArray[index] != "%") 
        caption += (displayUOM[instancesArray[index]] === undefined ? "" : displayUOM[instancesArray[index]])
     
//      if (chartType == 0 )
//        caption += (captionsArrayLength < 2 || index == captionsArrayLength - 1 ? 
//          (displayUOM[instancesArray[index]] === undefined ? "" : displayUOM[instancesArray[index]] ) : "" )
//      if (chartType == 1 ) {  // Have room for all UOMs
//        if(captionsArray[index] != "%") 
//          caption += (displayUOM[instancesArray[index]] === undefined ? "" : displayUOM[instancesArray[index]])
//      }
    }

    // Finish space  
    caption += (index < captionsArrayLength - 1 ? " " : "")

    // Remove 'undefined', 'NaN'
    return caption.replace("undefined",".").replace("NaN",".")

  }

  //
  // mainTimer: Read dbus for values and build PathLine(s) every graphInterval 
  //
  Timer
  {
    id:mainTimer
    interval: 1000      // Keep at 1000 msec, use counter to redraw graph lines
    running: true
    repeat: true
    onTriggered: {

      graphCounter += 1000


      if(!firstTime && graphCounter < graphInterval) return
      
      // First time through build set all values to current dbus value and
      // set actual graphInterval
      if(firstTime) {

        firstTime = false

        // Set my timer to graphInterval
        //mainTimer.interval = graphInterval

        // Initialize pathList[] with scaled x position
        var scaleXAxis = (chartFrame.width) / graphPoints
        
        for(var i=0; i< captionsArrayLength; i++) {
          for(var j=0; j < graphPoints; j++) {
            pathList[i].pathElements[j].x = scaleXAxis * j + chartFrame.border.width
          }
        }
      
      // Initialze curValYn
      copyCurrentValues()
      
      }   // firstTime

      // Display next refresh time
      refreshTime = Qt.formatDateTime(new Date(new Date().getTime() + graphInterval), "h:mm:ss")

      // Start cycle
      graphCounter = 0

      // Fetch the cached graph values from dbus and store in yArray
      // /Temperature values are in F
      var dbusSplit = (graphCache.value).split(",")

      // yArray is a 2 dimensional array y[_numberPaths][graphPoints]
      var yArray = []               // outside array, 1 - captionsArrayLength
      var tmpCurrentValues = [curValY0, curValY1, curValY2, curValY3, curValY4, curValY5, curValY6, curValY7, curValY8, curValY9]
      
      for(var i=0; i< captionsArrayLength; i++) {

        yArray.push([])             // push new empty array []

        for(var j=0; j< graphPoints; j++) {
          // Extend array and initialize to current dbus value
          //yArray[i].push((isValue(i) ? valueList[i].value : 0.0))
          yArray[i].push(tmpCurrentValues[i])
          
          if(dbusSplit.length > 1 && (i * graphPoints + j) < dbusSplit.length) {
            yArray[i][j] = parseFloat(dbusSplit[i * graphPoints + j])
          }
        }  // j

        // Every time this triggers, push down y[i] values
        for(var j = 1; j < graphPoints; j++) {
          yArray[i][j - 1] = yArray[i][j]
        }

        // Store current value
        // DM 20230727 -- Added valueList.valid
        // If dbus disconnected (e.g., Bluetooth), this will keep the last value constant
        yArray[i][graphPoints - 1] = tmpCurrentValues[i]

      }  // i

      // Write new dbus cache string
      writeDbusCache(yArray)
      
      // Initial values for yMin/yMax; these will be reset by scaleYAxis()
      yMin = yArray[0][0]
      yMax = yMin
      yAvg = yMin

      // Set scaled values for graph and copy into pathList[i].pathElements[j]
      var scaleY = scaleYAxis(yArray)
      for(var i=0; i< captionsArrayLength; i++) {
        for(var j=0; j< graphPoints; j++) {
          pathList[i].pathElements[j].y = scaleY * (yArray[i][j] - yMin) - 1 * chartFrame.border.width - 1
          // 20231213 - Sanity checks, when BT devices are offline, results are unpredictable
          if(pathList[i].pathElements[j].y > 0 || pathList[i].pathElements[j].y < -1 * chartFrame.height)
            pathList[i].pathElements[j].y = 0
        }
      }
    pZero = scaleY * (0 - yMin) - 1 * chartFrame.border.width - 1  
    }  // OnTriggered
  }  // Timer

  //*********** Graph Display View **************//
  
  transform: Scale {
  // transform: Scale { origin.x: 25; origin.y: 25; xScale: 3}
    id: tform
  }
  property int _zoom: -1
  
  // Graph's outside rectangle to draw chart
  Rectangle {
    id:       graphFrame
    color:    backgroundColor    
    border    {color: "#0e2a53";  width: 1 }
    anchors   {fill:parent; margins: 0 }
    radius:   1
    visible:  isVisible

    // Display graphId in upper left corner
    Text {
      text: graphId
      horizontalAlignment: Text.AlignLeft
      anchors { fill:parent; leftMargin:1}
      color: borderColor
      font.pixelSize: 6
      }

    // Graph line rectangle
    Rectangle {
      id: chartFrame
      color:  bodyColor
      border {color: borderColor; width: chartBorderWidth }

      // Whitespace around body of graph, room for bottom caption
      anchors {
        fill: parent
        topMargin: 5
        bottomMargin: (chartType == 0 ? 17 : 5)
        leftMargin: 6
        rightMargin: 6
      }

    // Information area
    Text {
      text: refreshTime
      horizontalAlignment: Text.AlignRight
      anchors { fill:parent;topMargin: 12; rightMargin: chartBorderWidth;}
      color: borderColor
      font.pixelSize: 8
      }

      // Draw tick-marks @ hour **** only if interval > ???? ***
      Repeater {
        model: (chartType == 0 ? ticks + 1 : 0)

        Line {
          lineWidth: 1
          color: borderColor
          anchors {
            bottom: parent.bottom
            bottomMargin: -2 
          }
          property Path tickPath: Path {
            PathLine {x: index * (chartFrame.width - chartFrame.border.width/2)/ticks; y:-3}
            PathLine {x: index * (chartFrame.width - chartFrame.border.width/2)/ticks; y:-1}
          }
          Component.onCompleted: {
            from = tickPath.pathElements[0]
            to = tickPath.pathElements[1]
          }
        }
      }

      // Using paths[], draw graph lines
      Repeater {
        model: (chartType == 0 ? captionsArrayLength : 0)
        Item {
          property int index1: index      // Number graphs
          anchors.bottom: parent.bottom

          Repeater {                      // Data points (48)
            model: graphPoints - 1        // Note subtract 1 since to 
                                          // point is next point, index2 + 1


            Line {                        // Victron Line.qml -- just a rotated rectangle!

              property int index2: index
//throws non-notifiable properties.... 

              lineWidth: chartLineWidth       // (Line->rectangle.height)

              // lineTypes: 0=solid, 1=dash, 50% color if inactive (Line->rectangle.color)
              color: (index2 % 2 == 0  & lineTypesArray[index1] == 1 ? bodyColor : 
//                getColor(index1, fadeColorPercent, 0) )
                getColor(index1, fadeColorPercent, pathList[index1].pathElements[index2]) )

              Component.onCompleted: {
                if(captionsArray[index1] != "" && captionsArray[index1] != "-") {  // DM Don't graph title "-"
                  from = pathList[index1].pathElements[index2]
                  to = pathList[index1].pathElements[index2 + 1]
                }
              }
            } // Line
          } // Repeater 0-graphPoints
        } // Item
      } // Repeater 0 - captionsArrayLength

      // Graph y axis scale values, yMin, yAvg, yMax
      Text {
        text: ((yMax - yMin > 0) ? yMax.toFixed(yAxisDecimals) : "")
        anchors {
          top:parent.top
          left:parent.left
          leftMargin: 2
        }
        color: borderColor
        font.pixelSize: 8
        visible: (chartType == 0)
        }
      Text {
        text: ((yMax - yMin > 2) ? yAvg.toFixed(yAxisDecimals) : "")
        anchors {
          verticalCenter:parent.verticalCenter
          left:parent.left
          leftMargin: 2
        }
        color: borderColor
        font.pixelSize: 8
        visible: (chartType == 0)
      }
      Text {
        text: yMin.toFixed(yAxisDecimals)
        anchors {
          bottom:parent.bottom
          left:parent.left
          leftMargin:2
        }
        color: borderColor
        font.pixelSize: 8
        visible: (chartType == 0)
      }
      Text {
        text: (chartType == 0 ? displayInterval() : (graphInterval/1000).toFixed(0) + "sec")
        anchors {
          // leftMargin:-10
          fill:parent
          topMargin: 2
          //top:parent.top
          //left:parent.right
        }
        horizontalAlignment: Text.AlignRight
        color: borderColor
        font.pixelSize: 8
      }
    }   // chartFrame

    // Display current dbus value(s)
    Row {
      visible: (chartType == 0)
      anchors {
        horizontalCenter: parent.horizontalCenter
        bottom: parent.bottom
        bottomMargin: 1
        //  bottomMargin: (captionsFontSize == 12 ? 0 : 6 - captionsFontSize / 2)
      }
      Repeater {
        model: captionsArrayLength
        Text {
          text: (chartType == 0 ? makeCaption(index) : "")
          color: getColor(index, fadeColorPercent, pathList[index].pathElements[graphPoints - 1] )
          font {pixelSize: captionsFontSize; bold: true }
        }
      }
    } 
    
    // Table view
    Column {
      visible: (chartType == 1)
      anchors {
        //horizontalCenter: chartFrame.horizontalCenter
        ////left: chartFrame.left + 100
        //top: chartFrame.top 
        fill: parent
        leftMargin: 10
        topMargin: 5
      }
      Repeater {
        model: captionsArrayLength
        Text {
          text: (chartType == 1 ? makeCaption(index) : "")
          color: getColor(index, fadeColorPercent, pathList[index].pathElements[graphPoints - 1])
          font { pixelSize: captionsFontSize; bold: true }
        }
      }
    }
  }
}