/*
 * PageSettingsIntuition1.qml - Maintain Intuition Graph Settings (Graph)
 *
 * Copyright © 2023,2024 IntuitionSailing.com.  All rights reserved.
 * Contact: IntiutionSailing http://www.intuitionsailing.com don@intuitionsailing.com
 *
*/

import QtQuick 1.1

MbPage {
  property string graphId
  property string myPath: "com.victronenergy.settings/Settings/Intuition/Graphs/" + graphId
  title: qsTr("Graph " + graphId + " Definition")

  model: VisualItemModel
  {
    // Type - type of graph: 0 = Line graph, 1 = Numeric display
    // Optional, defaults to graph.
    MbItemOptions {
      description: qsTr("Graph Type, default Graph")
      bind: myPath + "/Type"
      show: item.valid
      unknownOptionText: "Default Graph"
      possibleValues: [
        MbOption { description: qsTr("Graph"); value: "0" },
        MbOption { description: qsTr("Numeric Display"); value: "1" }
      ]
    }
    // Width - width of graph, when multiple graphs are combined onto a single page
    MbEditBox {
      description: qsTr ("Width, default 160px")
      maximumLength: 3
      item.bind: myPath + "/Width"
      matchString: "0123456789"
    }
    // Height - height of graph, when multiple graphs are combined onto a single page
    MbEditBox {
      description: qsTr ("Height, default 84")
      maximumLength: 3
      item.bind: myPath + "/Height"
      matchString: "0123456789"
    }
    // Border Color - outside border color
    // Optional, defaults #808080
    MbEditBox {
      description: qsTr ("Border Color, default #808080")
      maximumLength: 7
      item.bind: myPath + "/BorderColor"
      matchString: "#0123456789ABCDEFabcdef"
      //editFontPixelSize: 10
    }
    // Body Color - inside graph body color
    // Optional, defaults to #faf9f5
    MbEditBox {
      description: qsTr ("Body Color, default #faf9f5")
      maximumLength: 7
      item.bind: myPath + "/BodyColor"
      matchString: "#0123456789ABCDEFabcdef"
      //editFontPixelSize: 10
    }
    // Background Color - background color framing the graph
    // Optional, defaults to #d0d0d0
    MbEditBox {
      description: qsTr ("Background Color, default #d0d0d0")
      maximumLength: 7
      item.bind: myPath + "/BackgroundColor"
      matchString: "#0123456789ABCDEFabcdef"
      //editFontPixelSize: 10
    }
    // Time Interval - X scale, duration, hours
    // Defaults to 24, i.e., display 24 hour graph
    MbEditBox {
      description: qsTr ("Time Interval Hours, default 24")
      maximumLength: 3
      item.bind: myPath + "/Interval"
      matchString: "0123456789."
      //editFontPixelSize: 10
    }
    // Captions - comma separated list of captions, displayed on the
    // bottom of the graph.
    // Required. 
    // Special cases:
    //   ~ Display graph, no caption
    //   % Type 0 = display graph, no caption, no value, chart, Type 1 no caption, no value, no UOM
    //   * System /CustomName
    //   - Don't display graph, no caption, don't include totals
    MbEditBox {
      description: qsTr ("Captions, required,'*', '~', '%', '-', ','")
      maximumLength: 100
      item.bind: myPath + "/Captions"
      editFontPixelSize: 8
    }
    // Captions Font Size - font size of caption
    // Optional, defaults to 10px
    MbItemOptions {
      description: qsTr("Captions Font Size, default 10px")
      bind: myPath + "/CaptionsFontSize"
      show: item.valid
      unknownOptionText: "Default 10"
      possibleValues: [
        MbOption {description: qsTr("8px"); value: "8"},
        MbOption {description: qsTr("9px"); value: "9"},
        MbOption {description: qsTr("10px"); value: "10"},
        MbOption {description: qsTr("11px"); value: "11"},
        MbOption {description: qsTr("12px"); value: "12"},
        MbOption {description: qsTr("14px"); value: "14"},
        MbOption {description: qsTr("16px"); value: "16"}
      ]
    }
    // Services and Service Mappings - CSV list defines the dbus path to the
    // data element.  Path can be either a "Service" or indirect
    // through the system "ServiceMapping.  For example, the Ruuvi
    // tags setup a ServiceMapping 'com_victronenergy_temperature_20' 
    // maps to service 'com.victronenergy.temperature.ruuvi_f50d4682a759'.
    // Either a Service or ServieMapping is required.  This avoids using
    // USB paths.
    // Note, Services and ServiceMappings are separate setup values because
    // Victron code throws an error when the service is invalid.
    MbEditBox {
      description: qsTr ("Services, optional/required, ','")
      maximumLength: 250
      item.bind: myPath + "/Services"
      editFontPixelSize: 8
    }
    MbEditBox {
      description: qsTr ("Service Mappings, optional/required, ','")
      maximumLength: 250
      item.bind: myPath + "/ServiceMappings"
      editFontPixelSize: 8
    }
    // Instances - actual data value, e.g., /Temperature
    // Also, can define an RPN forumula on the data elements.  For
    // example, assume define instances as,
    //  /ConsumedAmphours,/Dc/Pv/Power,[1] [4] /,/Level,/Dc/0/Voltage
    //  [1] [4] / ==>  Dc/Pv/Power divided by /Dc/0/Voltage
    MbEditBox {
      description: qsTr ("Instances, required, ','")
      maximumLength: 250
      item.bind: myPath + "/Instances"
      editFontPixelSize: 8
    }
    // Line colors - CSV list of line colors for one or many graph lines.
    // Optional, defaults to "#e10000,#075f07,#0e2a53,#e100e1,#e17014,#00e1e1,#80000,#c0c0c0,#808080,#000000".
    // Also, can define different colors for positive vs. negative values.  For example,
    // with /Dc/Battery/Current,  #075f07|#e10000 sets green line for positive (generation) amps, red line 
    // for negative (consumed) amps.
    MbEditBox {
      description: qsTr ("Line Colors ',',[+]|[-]")
      maximumLength: 35
      item.bind: myPath + "/LineColors"
      matchString: "#0123456789ABCDEFabcdef|,"
      editFontPixelSize: 8
    }
    // Line types - CSV list of line types, defining type of line: 0=solid, 1=dash
    //
    MbEditBox { 
      description: qsTr ("Line Types, default '0,...', 0=Solid 1=Dash, ','")
      maximumLength: 10
      item.bind: myPath + "/LineTypes"
      matchString: "01,"
      //editFontPixelSize: 10
    }
    // Decimals - CSV list of number of decimal places to show on caption, scale
    // Optional, defaults to 0,0,0,0,0...
    MbEditBox {
      description: qsTr ("Decimals, default '0,...', ','")
      maximumLength: 25
      item.bind: myPath + "/Decimals"
      matchString: "0123456789,"
      //editFontPixelSize: 10
    }
    // YMin - override the Y axis scale.  This is useful for values with a fixed range,
    // such as /Humidity or /SOC, e.g., 0
    MbEditBox {
      description: qsTr ("Y Min Fixed, default Auto")
      maximumLength: 12
      item.bind: myPath + "/YMin"
      matchString: "0123456789.-"
      //editFontPixelSize: 10
    }
    // YMax - override the Y axis scale.  This is useful for values with a fixed range,
    // such as /Humidity or /SOC, e.g., 100.
    MbEditBox {
      description: qsTr ("Y Max Fixed, default Auto")
      maximumLength: 12
      item.bind: myPath + "/YMax"
      matchString: "0123456789.-"
      //editFontPixelSize: 10
    }
  }
}
